@@ -1,5 +1,7 @@
 require 'mongrel'
 require 'stringio'
+require 'rack/content_length'
+require 'rack/chunked'
 
 module Rack
   module Handler
@@ -33,7 +35,7 @@ module Rack
       end
 
       def initialize(app)
-        @app = app
+        @app = Rack::Chunked.new(Rack::ContentLength.new(app))
       end
 
       def process(request, response)
@@ -45,7 +47,7 @@ module Rack
 
         env.update({"rack.version" => [0,1],
                      "rack.input" => request.body || StringIO.new(""),
-                     "rack.errors" => STDERR,
+                     "rack.errors" => $stderr,
 
                      "rack.multithread" => true,
                      "rack.multiprocess" => false, # ???
@@ -63,7 +65,7 @@ module Rack
           response.send_status(nil)
 
           headers.each { |k, vs|
-            vs.each { |v|
+            vs.split("\n").each { |v|
               response.header[k] = v
             }
           }
