@@ -8,11 +8,23 @@ module Rack
   #   req = Rack::Request.new(env)
   #   req.post?
   #   req.params["data"]
+  #
+  # The environment hash passed will store a reference to the Request object
+  # instantiated so that it will only instantiate if an instance of the Request
+  # object doesn't already exist.
 
   class Request
     # The environment of the request.
     attr_reader :env
 
+    def self.new(env, *args)
+      if self == Rack::Request
+        env["rack.request"] ||= super
+      else
+        super
+      end
+    end
+
     def initialize(env)
       @env = env
     end
@@ -26,6 +38,8 @@ module Rack
     def query_string;    @env["QUERY_STRING"].to_s                end
     def content_length;  @env['CONTENT_LENGTH']                   end
     def content_type;    @env['CONTENT_TYPE']                     end
+    def session;         @env['rack.session'] ||= {}              end
+    def session_options; @env['rack.session.options'] ||= {}      end
 
     # The media type (type/subtype) portion of the CONTENT_TYPE header
     # without any media type parameters. e.g., when CONTENT_TYPE is
@@ -34,7 +48,7 @@ module Rack
     # For more information on the use of media types in HTTP, see:
     # http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
     def media_type
-      content_type && content_type.split(/\s*[;,]\s*/, 2)[0].downcase
+      content_type && content_type.split(/\s*[;,]\s*/, 2).first.downcase
     end
 
     # The media type parameters provided in CONTENT_TYPE as a Hash, or
@@ -80,6 +94,14 @@ module Rack
       'multipart/form-data'
     ]
 
+    # The set of media-types. Requests that do not indicate
+    # one of the media types presents in this list will not be eligible
+    # for param parsing like soap attachments or generic multiparts
+    PARSEABLE_DATA_MEDIA_TYPES = [
+      'multipart/related',
+      'multipart/mixed'
+    ]  
+
     # Determine whether the request body contains form-data by checking
     # the request media_type against registered form-data media-types:
     # "application/x-www-form-urlencoded" and "multipart/form-data". The
@@ -89,6 +111,12 @@ module Rack
       FORM_DATA_MEDIA_TYPES.include?(media_type)
     end
 
+    # Determine whether the request body contains data by checking
+    # the request media_type against registered parse-data media-types
+    def parseable_data?
+      PARSEABLE_DATA_MEDIA_TYPES.include?(media_type)
+    end
+
     # Returns the data recieved in the query string.
     def GET
       if @env["rack.request.query_string"] == query_string
@@ -96,7 +124,7 @@ module Rack
       else
         @env["rack.request.query_string"] = query_string
         @env["rack.request.query_hash"]   =
-          Utils.parse_query(query_string)
+          Utils.parse_nested_query(query_string)
       end
     end
 
@@ -107,13 +135,19 @@ module Rack
     def POST
       if @env["rack.request.form_input"].eql? @env["rack.input"]
         @env["rack.request.form_hash"]
-      elsif form_data?
+      elsif form_data? || parseable_data?
         @env["rack.request.form_input"] = @env["rack.input"]
         unless @env["rack.request.form_hash"] =
             Utils::Multipart.parse_multipart(env)
-          @env["rack.request.form_vars"] = @env["rack.input"].read
-          @env["rack.request.form_hash"] = Utils.parse_query(@env["rack.request.form_vars"])
-          @env["rack.input"].rewind if @env["rack.input"].respond_to?(:rewind)
+          form_vars = @env["rack.input"].read
+
+          # Fix for Safari Ajax postings that always append \0
+          form_vars.sub!(/\0\z/, '')
+
+          @env["rack.request.form_vars"] = form_vars
+          @env["rack.request.form_hash"] = Utils.parse_nested_query(form_vars)
+
+          @env["rack.input"].rewind
         end
         @env["rack.request.form_hash"]
       else
@@ -188,11 +222,13 @@ module Rack
 
       url
     end
-
+    
+    def path
+      script_name + path_info
+    end
+    
     def fullpath
-      path = script_name + path_info
-      path << "?" << query_string  unless query_string.empty?
-      path
+      query_string.empty? ? path : "#{path}?#{query_string}"
     end
 
     def accept_encoding
